#-----------------------------------------------------------------------------
# Copyright (c) 2012 - 2015, Anaconda, Inc. All rights reserved.
#
# Powered by the Bokeh Development Team.
#
# The full license is in the file LICENSE.txt, distributed with this software.
#-----------------------------------------------------------------------------

from __future__ import absolute_import
from mock import patch, Mock, PropertyMock
import os

from PIL import Image
import pytest
import selenium.webdriver as webdriver
import unittest

import bokeh.io as io

from bokeh.application.application import Application
from bokeh.resources import Resources
from bokeh.models.plots import Plot
from bokeh.models import Range1d

class TestDefaultState(unittest.TestCase):

    def test_type(self):
        self.assertTrue(isinstance(io._state, io.State))

class testCurdoc(unittest.TestCase):

    def test(self):
        self.assertEqual(io.curdoc(), io._state.document)

class testCurstate(unittest.TestCase):

    def test(self):
        self.assertEqual(io.curstate(), io._state)


class DefaultStateTester(unittest.TestCase):

    def _check_func_called(self, func, args, kwargs):
        self.assertTrue(func.called)
        self.assertEqual(func.call_args[0], args)
        self.assertEqual(func.call_args[1], kwargs)

    def setUp(self):
        self._orig_state = io._state
        io._state = Mock()
        doc = Mock()
        roots = PropertyMock(return_value=[])
        type(doc).roots = roots
        io._state.document = doc

    def tearDown(self):
        io._state = self._orig_state
        io._state.document.clear()

class testOutputFile(DefaultStateTester):

    def test_noarg(self):
        default_kwargs = dict(title="Bokeh Plot", mode="cdn", root_dir=None)
        io.output_file("foo.html")
        self._check_func_called(io._state.output_file, ("foo.html",), default_kwargs)

    def test_args(self):
        kwargs = dict(title="title", mode="cdn", root_dir="foo")
        io.output_file("foo.html", **kwargs)
        self._check_func_called(io._state.output_file, ("foo.html",), kwargs)

class TestOutputJupyter(DefaultStateTester):

    @patch('bokeh.io.load_notebook')
    def test_noarg(self, mock_load_notebook):
        # call _install_notebook_hook to register the patched `bokeh.io.load_notebook`
        from bokeh.io import _install_notebook_hook
        _install_notebook_hook()
        default_load_jupyter_args = (None, False, False, 5000)
        io.output_notebook()
        self._check_func_called(io._state.output_notebook, ('jupyter',), {})
        self._check_func_called(mock_load_notebook, default_load_jupyter_args, {'notebook_type': 'jupyter'})

    @patch('bokeh.io.load_notebook')
    def test_args(self, mock_load_notebook):
        # call _install_notebook_hook to register the patched `bokeh.io.load_notebook`
        from bokeh.io import _install_notebook_hook
        _install_notebook_hook()
        load_jupyter_args = (Resources(), True, True, 1000)
        io.output_notebook(*load_jupyter_args)
        self._check_func_called(io._state.output_notebook, ('jupyter',), {})
        self._check_func_called(mock_load_notebook, load_jupyter_args, {'notebook_type': 'jupyter'})

class TestSave(DefaultStateTester):
    pass

class Test_GetSaveArgs(DefaultStateTester):

    def test_explicit_filename(self):
        filename, resources, title = io._get_save_args(io._state, "filename", "resources", "title")
        self.assertEqual(filename, "filename")

    def test_default_filename(self):
        io._state.file = {}
        io._state.file['filename'] = "filename"
        filename, resources, title = io._get_save_args(io._state, None, "resources", "title")
        self.assertEqual(filename, "filename")

    def test_explicit_resources(self):
        filename, resources, title = io._get_save_args(io._state, "filename", "resources", "title")
        self.assertEqual(resources, "resources")

    def test_default_resources(self):
        io._state.file = {}
        io._state.file['resources'] = "resources"
        filename, resources, title = io._get_save_args(io._state, "filename", None, "title")
        self.assertEqual(resources, "resources")

    @patch('warnings.warn')
    def test_missing_resources(self, mock_warn):
        from bokeh.resources import CDN
        io._state.file = None
        filename, resources, title = io._get_save_args(io._state, "filename", None, "title")
        self.assertEqual(resources, CDN)
        self.assertTrue(mock_warn.called)
        self.assertEqual(mock_warn.call_args[0], ("save() called but no resources were supplied and output_file(...) "
                                                  "was never called, defaulting to resources.CDN",))
        self.assertEqual(mock_warn.call_args[1], {})

    def test_explicit_title(self):
        filename, resources, title = io._get_save_args(io._state, "filename", "resources", "title")
        self.assertEqual(title, "title")

    def test_default_title(self):
        io._state.file = {}
        io._state.file['title'] = "title"
        filename, resources, title = io._get_save_args(io._state, "filename", "resources", None)
        self.assertEqual(title, "title")

    @patch('warnings.warn')
    def test_missing_title(self, mock_warn):
        io._state.file = None
        filename, resources, title = io._get_save_args(io._state, "filename", "resources", None)
        self.assertEqual(title, "Bokeh Plot")
        self.assertTrue(mock_warn.called)
        self.assertEqual(mock_warn.call_args[0], ("save() called but no title was supplied and output_file(...) "
                                                  "was never called, using default title 'Bokeh Plot'",))
        self.assertEqual(mock_warn.call_args[1], {})

class Test_SaveHelper(DefaultStateTester):

    @patch('io.open')
    @patch('bokeh.io.file_html')
    def test_save_helper_method(self, mock_file_html, mock_io_open):
        obj = Plot()
        filename, resources, title = io._get_save_args(io._state, "filename", "resources", "title")

        io._save_helper(obj, filename, resources, title)

        self._check_func_called(mock_file_html,
                                (obj, resources),
                                {"title": "title"})
        self._check_func_called(mock_io_open,
                                (filename,),
                                {"mode":"w", "encoding":"utf-8"})

class TestShow(DefaultStateTester):

    @patch('bokeh.io._show_with_state')
    def test_default_args(self, mock__show_with_state):
        default_kwargs = dict(browser=None, new="tab", notebook_handle=False)
        io.show("obj", **default_kwargs)
        self._check_func_called(mock__show_with_state, ("obj", io._state, None, "tab"), {'notebook_handle': False})

    @patch('bokeh.io._show_with_state')
    def test_explicit_args(self, mock__show_with_state):
        default_kwargs = dict(browser="browser", new="new", notebook_handle=True)
        io.show("obj", **default_kwargs)
        self._check_func_called(mock__show_with_state, ("obj", io._state, "browser", "new"), {'notebook_handle': True})

    @patch('bokeh.io._run_notebook_hook')
    def test_show_with_app(self, mock__run_notebook_hook):
        app = Application()
        io.show(app, notebook_url="baz")
        io._state.notebook_type == "foo"
        self.assertTrue(mock__run_notebook_hook.called)
        self.assertEqual(mock__run_notebook_hook.call_args[0][0], io._state.notebook_type)
        self.assertEqual(mock__run_notebook_hook.call_args[0][1:], ("app", app, io._state, "baz"))
        self.assertEqual(mock__run_notebook_hook.call_args[1], {})

@patch('bokeh.io._show_with_state')
def test_show_adds_obj_to_document_if_not_already_there(m):
    assert io._state.document.roots == []
    p = Plot()
    io.show(p)
    assert p in io._state.document.roots


@patch('bokeh.io._show_with_state')
def test_show_doesnt_duplicate_if_already_there(m):
    io._state.document.clear()
    p = Plot()
    io.show(p)
    assert io._state.document.roots == [p]
    io.show(p)
    assert io._state.document.roots == [p]


class Test_ShowWithState(DefaultStateTester):

    @patch('bokeh.io._show_notebook_doc_with_state')
    @patch('bokeh.io._show_file_with_state')
    @patch('bokeh.util.browser.get_browser_controller')
    def test_notebook(self, mock_get_browser_controller,
            mock__show_file_with_state, mock__show_notebook_doc_with_state):
        mock_get_browser_controller.return_value = "controller"
        s = io.State()
        s.output_notebook()
        io._show_with_state("obj", s, "browser", "new")
        self._check_func_called(mock__show_notebook_doc_with_state, ("obj", s, False), {})
        self.assertFalse(mock__show_file_with_state.called)

        s.output_file("foo.html")
        io._show_with_state("obj", s, "browser", "new")
        self._check_func_called(mock__show_notebook_doc_with_state, ("obj", s, False), {})
        self._check_func_called(mock__show_file_with_state, ("obj", s, "new", "controller"), {})

    @patch('bokeh.io.get_comms')
    @patch('bokeh.io._show_notebook_doc_with_state')
    @patch('bokeh.io._show_file_with_state')
    @patch('bokeh.util.browser.get_browser_controller')
    def test_no_notebook(self, mock_get_browser_controller,
            mock__show_file_with_state, mock__show_notebook_doc_with_state,
            mock_get_comms):
        mock_get_browser_controller.return_value = "controller"
        mock_get_comms.return_value = "comms"
        s = io.State()

        s.output_file("foo.html")
        io._show_with_state("obj", s, "browser", "new")
        self.assertFalse(mock__show_notebook_doc_with_state.called)
        self._check_func_called(mock__show_file_with_state, ("obj", s, "new", "controller"), {})

class Test_ShowFileWithState(DefaultStateTester):

    @patch('os.path.abspath')
    @patch('bokeh.io.save')
    def test(self, mock_save, mock_abspath):
        s = io.State()
        s.output_file("foo.html")
        controller = Mock()
        mock_save.return_value = "savepath"

        io._show_file_with_state("obj", s, "window", controller)
        self._check_func_called(mock_save, ("obj",), {"state": s})
        self._check_func_called(controller.open, ("file://savepath",), {"new": 1})

        io._show_file_with_state("obj", s, "tab", controller)
        self._check_func_called(mock_save, ("obj",), {"state": s})
        self._check_func_called(controller.open, ("file://savepath",), {"new": 2})

class Test_ShowJupyterWithState(DefaultStateTester):

    @patch('bokeh.io.get_comms')
    @patch('bokeh.io.publish_display_data')
    @patch('bokeh.io.notebook_content')
    def test_no_server(self, mock_notebook_content, mock_publish_display_data, mock_get_comms):
        mock_get_comms.return_value = "comms"
        s = io.State()
        mock_notebook_content.return_value = ["notebook_script", "notebook_div"]

        class Obj(object):
            _id = None

        io._nb_loaded = True
        io._show_jupyter_doc_with_state(Obj(), s, True)
        io._nb_loaded = False

        expected_args = ({'application/javascript': 'notebook_script', 'application/vnd.bokehjs_exec.v0+json': ''},)
        expected_kwargs = {'metadata': {'application/vnd.bokehjs_exec.v0+json': {'id': None}}}

        self._check_func_called(mock_publish_display_data, expected_args, expected_kwargs)

class TestResetOutput(DefaultStateTester):

    def test(self):
        io.reset_output()
        self.assertTrue(io._state.reset.called)

def _test_layout_added_to_root(layout_generator, children=None):
    layout = layout_generator(Plot() if children is None else children)
    assert layout in io.curdoc().roots
    io.curdoc().clear()

def _test_children_removed_from_root(layout_generator, children=None):
    component = Plot()
    io.curdoc().add_root(component if children is None else children[0][0])
    layout_generator(component if children is None else children)
    assert component not in io.curdoc().roots
    io.curdoc().clear()

def test__crop_image():
    image = Image.new(mode="RGBA", size=(10,10))
    rect = dict(left=2, right=8, top=3, bottom=7)
    cropped = io._crop_image(image, **rect)
    assert cropped.size == (6,4)

def test__save_layout_html_resets_plot_dims():
    initial_height, initial_width = 200, 250

    layout = Plot(x_range=Range1d(), y_range=Range1d(),
                  plot_height=initial_height, plot_width=initial_width)

    io._save_layout_html(layout, height=100, width=100)
    assert layout.plot_height == initial_height
    assert layout.plot_width == initial_width

def test__notebook_hooks():
    assert "jupyter" in io._notebook_hooks
    # TODO (bev) can be improved after load_notebook is simplified
    assert "load" in io._notebook_hooks["jupyter"]
    assert io._notebook_hooks["jupyter"]['doc'] == io._show_jupyter_doc_with_state
    assert io._notebook_hooks["jupyter"]['app'] == io._show_jupyter_app_with_state

def test_install_notebook_hook():
    io.install_notebook_hook("foo", "load", "doc", "app")
    assert io._notebook_hooks["foo"]['load'] == "load"
    assert io._notebook_hooks["foo"]['doc'] == "doc"
    assert io._notebook_hooks["foo"]['app'] == "app"
    with pytest.raises(RuntimeError):
        io.install_notebook_hook("foo", "load2", "doc2", "app2")
    io.install_notebook_hook("foo", "load2", "doc2", "app2", overwrite=True)
    assert io._notebook_hooks["foo"]['load'] == "load2"
    assert io._notebook_hooks["foo"]['doc'] == "doc2"
    assert io._notebook_hooks["foo"]['app'] == "app2"

@pytest.mark.unit
@pytest.mark.selenium
def test__get_screenshot_as_png():
    layout = Plot(x_range=Range1d(), y_range=Range1d(),
                  plot_height=20, plot_width=20, toolbar_location=None,
                  outline_line_color=None, background_fill_color=None,
                  border_fill_color=None)

    png = io._get_screenshot_as_png(layout)
    assert png.size == (20, 20)
    # a 20x20px image of transparent pixels
    assert png.tobytes() == ("\x00"*1600).encode()

@pytest.mark.unit
@pytest.mark.selenium
def test__get_screenshot_as_png_with_driver():
    layout = Plot(x_range=Range1d(), y_range=Range1d(),
                  plot_height=20, plot_width=20, toolbar_location=None,
                  outline_line_color=None, background_fill_color=None,
                  border_fill_color=None)

    driver = webdriver.PhantomJS(service_log_path=os.path.devnull)

    png = io._get_screenshot_as_png(layout, driver=driver)

    # Have to manually clean up the driver session
    driver.quit()

    assert png.size == (20, 20)
    # a 20x20px image of transparent pixels
    assert png.tobytes() == ("\x00"*1600).encode()

@pytest.mark.unit
@pytest.mark.selenium
def test__get_screenshot_as_png_large_plot():
    layout = Plot(x_range=Range1d(), y_range=Range1d(),
                  plot_height=800, plot_width=800, toolbar_location=None,
                  outline_line_color=None, background_fill_color=None,
                  border_fill_color=None)

    driver = webdriver.PhantomJS(service_log_path=os.path.devnull)
    assert driver.get_window_size() == {'width': 400, 'height': 300}

    io._get_screenshot_as_png(layout, driver=driver)

    # LC: Although the window size doesn't match the plot dimensions (unclear
    # why), the window resize allows for the whole plot to be captured
    assert driver.get_window_size() == {'width': 1366, 'height': 768}

    # Have to manually clean up the driver session
    driver.quit()

@pytest.mark.unit
@pytest.mark.selenium
def test__get_svgs_no_svg_present():
    layout = Plot(x_range=Range1d(), y_range=Range1d(),
              plot_height=20, plot_width=20, toolbar_location=None)

    svgs = io._get_svgs(layout)
    assert svgs == []

@pytest.mark.unit
@pytest.mark.selenium
def test__get_svgs_with_svg_present():
    layout = Plot(x_range=Range1d(), y_range=Range1d(),
                  plot_height=20, plot_width=20, toolbar_location=None,
                  outline_line_color=None, border_fill_color=None,
                  background_fill_color=None, output_backend="svg")

    svgs = io._get_svgs(layout)
    assert svgs[0] == ('<svg version="1.1" xmlns="http://www.w3.org/2000/svg" xmlns:xlink="http://www.w3.org/1999/xlink" '
                       'width="20" height="20" style="width: 20px; height: 20px;"><defs/><g><g/><g transform="scale(1,1) '
                       'translate(0.5,0.5)"><rect fill="#FFFFFF" stroke="none" x="0" y="0" width="20" height="20"/><g/><g/><g/><g/></g></g></svg>')

@pytest.mark.unit
@pytest.mark.selenium
def test__get_svgs_with_svg_present_with_driver():
    layout = Plot(x_range=Range1d(), y_range=Range1d(),
                  plot_height=20, plot_width=20, toolbar_location=None,
                  outline_line_color=None, border_fill_color=None,
                  background_fill_color=None, output_backend="svg")

    driver = webdriver.PhantomJS(service_log_path=os.path.devnull)

    svgs = io._get_svgs(layout)

    # Have to manually clean up the driver session
    driver.quit()

    assert svgs[0] == ('<svg version="1.1" xmlns="http://www.w3.org/2000/svg" xmlns:xlink="http://www.w3.org/1999/xlink" '
                       'width="20" height="20" style="width: 20px; height: 20px;"><defs/><g><g/><g transform="scale(1,1) '
                       'translate(0.5,0.5)"><rect fill="#FFFFFF" stroke="none" x="0" y="0" width="20" height="20"/><g/><g/><g/><g/></g></g></svg>')

@patch('os.access')
def test__no_access(mock_access):
    import os
    io._no_access("test")
    assert mock_access.called
    assert mock_access.call_args[0] == ("test", os.W_OK | os.X_OK)
    assert mock_access.call_args[1] == {}

@patch('tempfile.NamedTemporaryFile')
def test__temp_filename(mock_tmp):
    fn = Mock()
    type(fn).name = PropertyMock(return_value="Junk.test")
    mock_tmp.return_value = fn

    r = io._temp_filename("test")
    assert r == "Junk.test"
    assert mock_tmp.called
    assert mock_tmp.call_args[0] == ()
    assert mock_tmp.call_args[1] == {'suffix': '.test'}

def test__shares_exec_prefix():
    import sys
    old_ex = sys.exec_prefix
    try:
        sys.exec_prefix = "/foo/bar"
        assert io._shares_exec_prefix("/foo/bar") == True
        sys.exec_prefix = "/baz/bar"
        assert io._shares_exec_prefix("/foo/bar") == False
        sys.exec_prefix = None
        assert io._shares_exec_prefix("/foo/bar") == False
    finally:
        sys.exec_prefix = old_ex

def test__detect_current_filename():
    assert io._detect_current_filename().endswith("py.test")

def test_default_filename():
    old__detect_current_filename = io._detect_current_filename
    old__no_access = io._no_access
    old__shares_exec_prefix = io._shares_exec_prefix

    io._detect_current_filename = lambda : "/a/b/foo.py"

    try:
        # .py extension
        with pytest.raises(RuntimeError):
            io.default_filename("py")

        # a current file, access, and no share exec
        io._no_access = lambda x: False
        r = io.default_filename("test")
        assert r == "/a/b/foo.test"

        # a current file, NO access, and no share exec
        io._no_access = lambda x: True
        r = io.default_filename("test")
        assert r != "/a/b/foo.test"
        assert r.endswith(".test")

        # a current file, access, but WITH share exec
        io._no_access = lambda x: False
        io._shares_exec_prefix = lambda x: True
        r = io.default_filename("test")
        assert r != "/a/b/foo.test"
        assert r.endswith(".test")

        # no current file
        io._detect_current_filename = lambda : None
        io._no_access = lambda x: False
        io._shares_exec_prefix = lambda x: False
        r = io.default_filename("test")
        assert r != "/a/b/foo.test"
        assert r.endswith(".test")

    finally:
        io._detect_current_filename = old__detect_current_filename
        io._no_access = old__no_access
        io._shares_exec_prefix = old__shares_exec_prefix



import bokeh.transform as bt

from bokeh.models import CategoricalColorMapper, Dodge, FactorRange, Jitter, LinearColorMapper, LogColorMapper

def test_transform():
    t = bt.transform("foo", "junk")
    assert t == dict(field="foo", transform="junk")

def test_dodge():
    t = bt.dodge("foo", 0.5)
    assert isinstance(t, dict)
    assert set(t) == {"field", "transform"}
    assert t['field'] == "foo"
    assert isinstance(t['transform'], Dodge)
    assert t['transform'].value == 0.5
    assert t['transform'].range is None

def test_dodge_with_range():
    r = FactorRange("a")
    t = bt.dodge("foo", 0.5, range=r)
    assert isinstance(t, dict)
    assert set(t) == {"field", "transform"}
    assert t['field'] == "foo"
    assert isinstance(t['transform'], Dodge)
    assert t['transform'].value == 0.5
    assert t['transform'].range is r
    assert t['transform'].range.factors == ["a"]

def test_jitter_defaults():
    t = bt.jitter("foo", width=0.5)
    assert isinstance(t, dict)
    assert set(t) == {"field", "transform"}
    assert t['field'] == "foo"
    assert isinstance(t['transform'], Jitter)
    assert t['transform'].width == 0.5
    assert t['transform'].mean == 0
    assert t['transform'].distribution == "uniform"
    assert t['transform'].range is None

def test_jitter():
    t = bt.jitter("foo", width=0.5, mean=0.1, distribution="normal")
    assert isinstance(t, dict)
    assert set(t) == {"field", "transform"}
    assert t['field'] == "foo"
    assert isinstance(t['transform'], Jitter)
    assert t['transform'].width == 0.5
    assert t['transform'].mean == 0.1
    assert t['transform'].distribution == "normal"
    assert t['transform'].range is None

def test_jitter_with_range():
    r = FactorRange("a")
    t = bt.jitter("foo", width=0.5, mean=0.1, range=r)
    assert isinstance(t, dict)
    assert set(t) == {"field", "transform"}
    assert t['field'] == "foo"
    assert isinstance(t['transform'], Jitter)
    assert t['transform'].width == 0.5
    assert t['transform'].mean == 0.1
    assert t['transform'].distribution == "uniform"
    assert t['transform'].range is r
    assert t['transform'].range.factors == ["a"]

def test_factor_cmap_defaults():
    t = bt.factor_cmap("foo", ["red", "green"], ["foo", "bar"])
    assert isinstance(t, dict)
    assert set(t) == {"field", "transform"}
    assert t['field'] == "foo"
    assert isinstance(t['transform'], CategoricalColorMapper)
    assert t['transform'].palette == ["red", "green"]
    assert t['transform'].factors == ["foo", "bar"]
    assert t['transform'].start == 0
    assert t['transform'].end is None
    assert t['transform'].nan_color == "gray"

def test_factor_cmap():
    t = bt.factor_cmap("foo", ["red", "green"], ["foo", "bar"], start=1, end=2, nan_color="pink")
    assert isinstance(t, dict)
    assert set(t) == {"field", "transform"}
    assert t['field'] == "foo"
    assert isinstance(t['transform'], CategoricalColorMapper)
    assert t['transform'].palette == ["red", "green"]
    assert t['transform'].factors == ["foo", "bar"]
    assert t['transform'].start == 1
    assert t['transform'].end is 2
    assert t['transform'].nan_color == "pink"

def test_linear_cmap_defaults():
    t = bt.linear_cmap("foo", ["red", "green"], 0, 10)
    assert isinstance(t, dict)
    assert set(t) == {"field", "transform"}
    assert t['field'] == "foo"
    assert isinstance(t['transform'], LinearColorMapper)
    assert t['transform'].palette == ["red", "green"]
    assert t['transform'].low == 0
    assert t['transform'].high is 10
    assert t['transform'].low_color is None
    assert t['transform'].high_color is None
    assert t['transform'].nan_color == "gray"

def test_linear_cmap():
    t = bt.linear_cmap("foo", ["red", "green"], 0, 10, low_color="orange", high_color="blue", nan_color="pink")
    assert isinstance(t, dict)
    assert set(t) == {"field", "transform"}
    assert t['field'] == "foo"
    assert isinstance(t['transform'], LinearColorMapper)
    assert t['transform'].palette == ["red", "green"]
    assert t['transform'].low == 0
    assert t['transform'].high is 10
    assert t['transform'].low_color == "orange"
    assert t['transform'].high_color == "blue"
    assert t['transform'].nan_color == "pink"

def test_log_cmap_defaults():
    t = bt.log_cmap("foo", ["red", "green"], 0, 10)
    assert isinstance(t, dict)
    assert set(t) == {"field", "transform"}
    assert t['field'] == "foo"
    assert isinstance(t['transform'], LogColorMapper)
    assert t['transform'].palette == ["red", "green"]
    assert t['transform'].low == 0
    assert t['transform'].high is 10
    assert t['transform'].low_color is None
    assert t['transform'].high_color is None
    assert t['transform'].nan_color == "gray"

def test_log_cmap():
    t = bt.log_cmap("foo", ["red", "green"], 0, 10, low_color="orange", high_color="blue", nan_color="pink")
    assert isinstance(t, dict)
    assert set(t) == {"field", "transform"}
    assert t['field'] == "foo"
    assert isinstance(t['transform'], LogColorMapper)
    assert t['transform'].palette == ["red", "green"]
    assert t['transform'].low == 0
    assert t['transform'].high is 10
    assert t['transform'].low_color == "orange"
    assert t['transform'].high_color == "blue"
    assert t['transform'].nan_color == "pink"



import sys
import subprocess
import bokeh


def test_main_module():
    """ Test the __main__.py that allow doing "python -m bokeh"
    """
    cmd = [sys.executable, '-m', 'bokeh', '--version']
    v = subprocess.check_output(cmd, stderr=subprocess.STDOUT).decode().strip()
    assert v == bokeh.__version__



''' Provides client API for connecting to a Bokeh server from a Python
process.

The primary uses for the ``bokeh.client`` are:

* Implementing testing infrastructure around Bokeh applications
* Creating and customizing specific sessions of a Bokeh application
  running *in a Bokeh Server*, before passing them to a viewer.

While it is also possible to run Bokeh application code "outside" a
Bokeh server using ``bokeh.client``, this practice is **HIGHLY DISCOURCED**.


'''
from __future__ import absolute_import

from .session import ClientSession, pull_session, push_session, show_session, DEFAULT_SESSION_ID



from numpy.testing import assert_allclose

import bokeh.driving as driving

def _collector(results):
    def foo(val):
        results.append(val)
    return foo

w = 0.3
A = 3
phi = 0.1
offset = 2

def test_sine_sequence_values():
    results = []
    func = driving.sine(w, A, phi, offset)(_collector(results))
    for i in range(4):
        func()
    assert_allclose(
        results,
        [2.2995002499404844, 3.1682550269259515, 3.932653061713073, 4.524412954423689]
    )

def test_cosine_sequence_values():
    results = []
    func = driving.cosine(w, A, phi, offset)(_collector(results))
    for i in range(4):
        func()
    assert_allclose(
        results,
        [4.985012495834077, 4.763182982008655, 4.294526561853465, 3.6209069176044197]
    )

def test_linear_sequence_values():
    results = []
    func = driving.linear(m=2.5, b=3.7)(_collector(results))
    for i in range(4):
        func()
    assert_allclose(results, [3.7, 6.2, 8.7, 11.2])

def test_bounce_sequence_values():
    results = []
    func = driving.bounce([0, 1, 5, -1])(_collector(results))
    for i in range(8):
        func()
    assert results == [0, 1, 5, -1, -1, 5, 1, 0]

def test_repeat_sequence_values():
    results = []
    func = driving.repeat([0, 1, 5, -1])(_collector(results))
    for i in range(8):
        func()
    assert results == [0, 1, 5, -1, 0, 1, 5, -1]

def test_count_sequence_values():
    results = []
    func = driving.count()(_collector(results))
    for i in range(8):
        func()
    assert results == list(range(8))



from __future__ import absolute_import

import mock
import unittest

import bs4

from bokeh.core.templates import DOC_JS, PLOT_DIV, NOTEBOOK_DIV
from bokeh.core.json_encoder import serialize_json

import bokeh.embed as embed
from bokeh.plotting import figure, curdoc
from bokeh.resources import CDN, JSResources, CSSResources
from bokeh.util.string import encode_utf8
from jinja2 import Template
from six import string_types

_embed_test_plot = None


def setUpModule():
    global _embed_test_plot
    _embed_test_plot = figure()
    _embed_test_plot.circle([1, 2], [2, 3])


def _stable_id():
    return 'ID'


def test_components_return_type():
    plot1 = figure()
    plot2 = figure()
    # This is a testing artefact, users dont' have to do this in practice
    curdoc().add_root(plot1)
    curdoc().add_root(plot2)

    r = embed.components(plot1)
    assert len(r) == 2

    _, divs = embed.components((plot1, plot2))
    assert isinstance(divs, tuple)

    _, divs = embed.components([plot1, plot2])
    assert isinstance(divs, tuple)

    _, divs = embed.components({"Plot 1": plot1, "Plot 2": plot2})
    assert isinstance(divs, dict)
    assert all(isinstance(x, string_types) for x in divs.keys())


@mock.patch('bokeh.embed.make_id', new_callable=lambda: _stable_id)
def test_plot_dict_returned_when_wrap_plot_info_is_false(mock_make_id):
    plot1 = figure()
    plot2 = figure()
    # This is a testing artefact, users dont' have to do this in practice
    curdoc().add_root(plot1)
    curdoc().add_root(plot2)

    expected_plotdict_1 = {"modelid": plot1.ref["id"], "elementid": "ID", "docid": "ID"}
    expected_plotdict_2 = {"modelid": plot2.ref["id"], "elementid": "ID", "docid": "ID"}

    _, plotdict = embed.components(plot1, wrap_plot_info=False)
    assert plotdict == expected_plotdict_1

    _, plotids = embed.components((plot1, plot2), wrap_plot_info=False)
    assert plotids == (expected_plotdict_1, expected_plotdict_2)

    _, plotiddict = embed.components({'p1': plot1, 'p2': plot2}, wrap_plot_info=False)
    assert plotiddict == {'p1': expected_plotdict_1, 'p2': expected_plotdict_2}



class TestComponents(unittest.TestCase):


    def test_result_attrs(self):
        script, div = embed.components(_embed_test_plot)
        html = bs4.BeautifulSoup(script, "lxml")
        scripts = html.findAll(name='script')
        self.assertEqual(len(scripts), 1)
        self.assertTrue(scripts[0].attrs, {'type': 'text/javascript'})

    def test_div_attrs(self):
        script, div = embed.components(_embed_test_plot)
        html = bs4.BeautifulSoup(div, "lxml")

        divs = html.findAll(name='div')
        self.assertEqual(len(divs), 2)

        div = divs[0]
        self.assertEqual(set(div.attrs), set(['class']))
        self.assertEqual(div.attrs['class'], ['bk-root'])
        self.assertEqual(div.text, '\n\n')

        div = divs[1]
        self.assertEqual(set(div.attrs), set(['id', 'class']))
        self.assertEqual(div.attrs['class'], ['bk-plotdiv'])
        self.assertEqual(div.text, '')

    def test_script_is_utf8_encoded(self):
        script, div = embed.components(_embed_test_plot)
        self.assertTrue(isinstance(script, str))

    @mock.patch('bokeh.embed.make_id', new_callable=lambda: _stable_id)
    def test_output_is_without_script_tag_when_wrap_script_is_false(self, mock_make_id):
        script, div = embed.components(_embed_test_plot)
        html = bs4.BeautifulSoup(script, "lxml")
        scripts = html.findAll(name='script')
        self.assertEqual(len(scripts), 1)
        script_content = scripts[0].getText()

        rawscript, div = embed.components(_embed_test_plot, wrap_script=False)
        self.maxDiff = None
        self.assertEqual(rawscript.strip(), script_content.strip())

class TestNotebookContent(unittest.TestCase):

    @mock.patch('bokeh.embed._standalone_docs_json_and_render_items')
    def test_notebook_content(self, mock_standalone_docs_json_and_render_items):
        (docs_json, render_items) = ("DOC_JSON", [dict(docid="foo", elementid="bar", modelid="bat")])
        mock_standalone_docs_json_and_render_items.return_value = (docs_json, render_items)

        expected_script = DOC_JS.render(docs_json=serialize_json(docs_json),
                                        render_items=serialize_json(render_items))
        expected_div = PLOT_DIV.render(elementid=render_items[0]['elementid'])

        (script, div) = embed.notebook_content(_embed_test_plot)

        self.assertEqual(script, expected_script)
        self.assertEqual(div, expected_div)

    @mock.patch('bokeh.embed._standalone_docs_json_and_render_items')
    def test_notebook_content_with_notebook_comms_target(self, mock_standalone_docs_json_and_render_items):
        (docs_json, render_items) = ("DOC_JSON", [dict(docid="foo", elementid="bar", modelid="bat")])
        mock_standalone_docs_json_and_render_items.return_value = (docs_json, render_items)
        comms_target = "NOTEBOOK_COMMS_TARGET"

        ## assert that NOTEBOOK_COMMS_TARGET is added to render_items bundle
        self.assertTrue('notebook_comms_target' not in render_items[0])
        (script, _) = embed.notebook_content(_embed_test_plot, notebook_comms_target=comms_target)
        self.assertTrue('notebook_comms_target' in render_items[0])

        ## assert that NOTEBOOK_COMMS_TARGET ends up in generated script
        expected_script = DOC_JS.render(docs_json=serialize_json(docs_json),
                                        render_items=serialize_json(render_items))

        self.assertEqual(script, expected_script)

class TestNotebookDiv(unittest.TestCase):

    @mock.patch('bokeh.embed.notebook_content')
    def test_notebook_div(self, mock_notebook_content):
        (script, div) = ("(function(){})()", "<div></div>")
        mock_notebook_content.return_value = (script, div)

        notebook_div = embed.notebook_div(_embed_test_plot)

        expected_notebook_div = NOTEBOOK_DIV.render(plot_script=script, plot_div=div)

        self.assertEqual(notebook_div, expected_notebook_div)

class TestFileHTML(unittest.TestCase):

    def test_return_type(self):

        class fake_template:
            def __init__(self, tester, user_template_variables=None):
                self.tester = tester
                self.template_variables = {
                    "title",
                    "bokeh_js",
                    "bokeh_css",
                    "plot_script",
                    "plot_div"
                }
                if user_template_variables is not None:
                    self.template_variables.update(user_template_variables)

            def render(self, template_variables):
                self.tester.assertTrue(
                    self.template_variables.issubset(
                        set(template_variables.keys())
                    )
                )
                return "template result"

        r = embed.file_html(_embed_test_plot, CDN, "title")
        self.assertTrue(isinstance(r, str))

        r = embed.file_html(_embed_test_plot, CDN, "title", fake_template(self))
        self.assertTrue(isinstance(r, str))

        r = embed.file_html(_embed_test_plot, CDN, "title",
                            fake_template(self, {"test_var"}),
                            {"test_var": "test"})
        self.assertTrue(isinstance(r, str))


def test_file_html_handles_js_only_resources():
    js_resources = JSResources(mode="relative", components=["bokeh"])
    template = Template("<head>{{ bokeh_js }}</head><body></body>")
    output = embed.file_html(_embed_test_plot, (js_resources, None), "title", template=template)
    html = encode_utf8("<head>%s</head><body></body>" % js_resources.render_js())
    assert output == html


@mock.patch('bokeh.embed.warn')
def test_file_html_provides_warning_if_no_css(mock_warn):
    js_resources = JSResources()
    embed.file_html(_embed_test_plot, (js_resources, None), "title")
    mock_warn.assert_called_once_with(
        'No Bokeh CSS Resources provided to template. If required you will need to provide them manually.'
    )


def test_file_html_handles_css_only_resources():
    css_resources = CSSResources(mode="relative", components=["bokeh"])
    template = Template("<head>{{ bokeh_css }}</head><body></body>")
    output = embed.file_html(_embed_test_plot, (None, css_resources), "title", template=template)
    html = encode_utf8("<head>%s</head><body></body>" % css_resources.render_css())
    assert output == html


@mock.patch('bokeh.embed.warn')
def test_file_html_provides_warning_if_no_js(mock_warn):
    css_resources = CSSResources()
    embed.file_html(_embed_test_plot, (None, css_resources), "title")
    mock_warn.assert_called_once_with(
        'No Bokeh JS Resources provided to template. If required you will need to provide them manually.'
    )


def test_file_html_title_is_escaped():
    r = embed.file_html(_embed_test_plot, CDN, "&<")
    assert "<title>&amp;&lt;</title>" in r


class TestAutoloadStatic(unittest.TestCase):

    def test_return_type(self):
        r = embed.autoload_static(_embed_test_plot, CDN, "some/path")
        self.assertEqual(len(r), 2)

    @mock.patch('bokeh.embed.make_id', new_callable=lambda: _stable_id)
    def test_script_attrs(self, mock_make_id):
        js, tag = embed.autoload_static(_embed_test_plot, CDN, "some/path")
        html = bs4.BeautifulSoup(tag, "lxml")
        scripts = html.findAll(name='script')
        self.assertEqual(len(scripts), 1)
        attrs = scripts[0].attrs
        self.assertTrue(set(attrs), set(['src',
            'data-bokeh-model-id',
            'id',
            'data-bokeh-doc-id']))
        self.assertEqual(attrs['data-bokeh-doc-id'], 'ID')
        self.assertEqual(attrs['data-bokeh-model-id'], str(_embed_test_plot._id))
        self.assertEqual(attrs['src'], 'some/path')


class TestConnectSessionOrDocument(unittest.TestCase):

    def test_invalid_resources_param(self):
        with self.assertRaises(ValueError):
            embed._connect_session_or_document(url="http://localhost:8081/foo/bar/sliders", resources=123)
        with self.assertRaises(ValueError):
            embed._connect_session_or_document(url="http://localhost:8081/foo/bar/sliders", resources="whatever")

    def test_resources_default_is_implicit(self):
        r = embed._connect_session_or_document(url="http://localhost:8081/foo/bar/sliders", resources="default")
        self.assertFalse('resources=' in r)

    def test_resources_none(self):
        r = embed._connect_session_or_document(url="http://localhost:8081/foo/bar/sliders", resources=None)
        self.assertTrue('resources=none' in r)


class TestServerDocument(unittest.TestCase):

    def test_ensure_no_session_do_model(self):
        r = embed.server_document(url="http://localhost:8081/foo/bar/sliders")
        self.assertTrue('bokeh-app-path=/foo/bar/sliders' in r)
        self.assertTrue('bokeh-absolute-url=http://localhost:8081/foo/bar/sliders' in r)
        html = bs4.BeautifulSoup(r, "lxml")
        scripts = html.findAll(name='script')
        self.assertEqual(len(scripts), 1)
        attrs = scripts[0].attrs
        self.assertTrue(set(attrs), set([
            'src',
            'data-bokeh-doc-id',
            'data-bokeh-model-id',
            'id'
        ]))
        divid = attrs['id']
        src = "%s/autoload.js?bokeh-autoload-element=%s&bokeh-app-path=/foo/bar/sliders&bokeh-absolute-url=%s" % \
              ("http://localhost:8081/foo/bar/sliders", divid, "http://localhost:8081/foo/bar/sliders")
        self.assertDictEqual({ 'data-bokeh-doc-id' : '',
                               'data-bokeh-model-id' : '',
                               'id' : divid,
                               'src' : src },
                             attrs)

class TestServerSession(unittest.TestCase):

    def test_model_and_session_both_required(self):
        with self.assertRaises(TypeError):
            embed.server_session()
        with self.assertRaises(TypeError):
            embed.server_session(_embed_test_plot)
        with self.assertRaises(TypeError):
            embed.server_session(session_id='fakesession')

    def test_ensure_session_and_model(self):
        r = embed.server_session(_embed_test_plot, session_id='fakesession')
        self.assertTrue('bokeh-session-id=fakesession' in r)
        html = bs4.BeautifulSoup(r, "lxml")
        scripts = html.findAll(name='script')
        self.assertEqual(len(scripts), 1)
        attrs = scripts[0].attrs
        self.assertTrue(set(attrs), set([
            'src',
            'data-bokeh-doc-id',
            'data-bokeh-model-id',
            'id'
        ]))
        divid = attrs['id']
        src = "%s/autoload.js?bokeh-autoload-element=%s&bokeh-absolute-url=%s&bokeh-session-id=fakesession" % \
              ("http://localhost:5006", divid, "http://localhost:5006")
        self.assertDictEqual({ 'data-bokeh-doc-id' : '',
                               'data-bokeh-model-id' : str(_embed_test_plot._id),
                               'id' : divid,
                               'src' : src },
                             attrs)

class TestAutoloadServer(unittest.TestCase):

    def test_return_type(self):
        r = embed.autoload_server(_embed_test_plot, session_id='fakesession')
        self.assertTrue(isinstance(r, str))

    def test_script_attrs_session_id_provided(self):
        r = embed.autoload_server(_embed_test_plot, session_id='fakesession')
        self.assertTrue('bokeh-session-id=fakesession' in r)
        html = bs4.BeautifulSoup(r, "lxml")
        scripts = html.findAll(name='script')
        self.assertEqual(len(scripts), 1)
        attrs = scripts[0].attrs
        self.assertTrue(set(attrs), set([
            'src',
            'data-bokeh-doc-id',
            'data-bokeh-model-id',
            'id'
        ]))
        divid = attrs['id']
        src = "%s/autoload.js?bokeh-autoload-element=%s&bokeh-absolute-url=%s&bokeh-session-id=fakesession" % \
              ("http://localhost:5006", divid, "http://localhost:5006")
        self.assertDictEqual({ 'data-bokeh-doc-id' : '',
                               'data-bokeh-model-id' : str(_embed_test_plot._id),
                               'id' : divid,
                               'src' : src },
                             attrs)

    def test_script_attrs_no_session_id_provided(self):
        r = embed.autoload_server(None)
        self.assertFalse('bokeh-session-id' in r)
        html = bs4.BeautifulSoup(r, "lxml")
        scripts = html.findAll(name='script')
        self.assertEqual(len(scripts), 1)
        attrs = scripts[0].attrs
        self.assertTrue(set(attrs), set([
            'src',
            'data-bokeh-doc-id',
            'data-bokeh-model-id',
            'id'
        ]))
        divid = attrs['id']
        src = "%s/autoload.js?bokeh-autoload-element=%s&bokeh-absolute-url=%s" % \
              ("http://localhost:5006", divid, "http://localhost:5006")
        self.assertDictEqual({ 'data-bokeh-doc-id' : '',
                               'data-bokeh-model-id' : '',
                               'id' : divid,
                               'src' : src },
                             attrs)

    def test_script_attrs_url_provided(self):
        r = embed.autoload_server(url="http://localhost:8081/foo/bar/sliders", relative_urls=True)
        self.assertTrue('bokeh-app-path=/foo/bar/sliders' in r)
        html = bs4.BeautifulSoup(r, "lxml")
        scripts = html.findAll(name='script')
        self.assertEqual(len(scripts), 1)
        attrs = scripts[0].attrs
        self.assertTrue(set(attrs), set([
            'src',
            'data-bokeh-doc-id',
            'data-bokeh-model-id',
            'id'
        ]))
        divid = attrs['id']
        src = "%s/autoload.js?bokeh-autoload-element=%s&bokeh-app-path=/foo/bar/sliders" % \
              ("http://localhost:8081/foo/bar/sliders", divid)
        self.assertDictEqual({ 'data-bokeh-doc-id' : '',
                               'data-bokeh-model-id' : '',
                               'id' : divid,
                               'src' : src },
                             attrs)

    def test_script_attrs_url_provided_absolute_resources(self):
        r = embed.autoload_server(url="http://localhost:8081/foo/bar/sliders")
        self.assertTrue('bokeh-app-path=/foo/bar/sliders' in r)
        self.assertTrue('bokeh-absolute-url=http://localhost:8081/foo/bar/sliders' in r)
        html = bs4.BeautifulSoup(r, "lxml")
        scripts = html.findAll(name='script')
        self.assertEqual(len(scripts), 1)
        attrs = scripts[0].attrs
        self.assertTrue(set(attrs), set([
            'src',
            'data-bokeh-doc-id',
            'data-bokeh-model-id',
            'id'
        ]))
        divid = attrs['id']
        src = "%s/autoload.js?bokeh-autoload-element=%s&bokeh-app-path=/foo/bar/sliders&bokeh-absolute-url=%s" % \
              ("http://localhost:8081/foo/bar/sliders", divid, "http://localhost:8081/foo/bar/sliders")
        self.assertDictEqual({ 'data-bokeh-doc-id' : '',
                               'data-bokeh-model-id' : '',
                               'id' : divid,
                               'src' : src },
                             attrs)

    def test_script_attrs_url_and_app_path_provided(self):
        for path in ("/foo/bar/sliders", "/foo/bar/sliders/", "foo/bar/sliders", "foo/bar/sliders"):
            r = embed.autoload_server(url="http://localhost:8081", app_path=path, relative_urls=True)
            self.assertTrue('bokeh-app-path=/foo/bar/sliders' in r)
            html = bs4.BeautifulSoup(r, "lxml")
            scripts = html.findAll(name='script')
            self.assertEqual(len(scripts), 1)
            attrs = scripts[0].attrs
            self.assertTrue(set(attrs), set([
                'src',
                'data-bokeh-doc-id',
                'data-bokeh-model-id',
                'id'
            ]))
            divid = attrs['id']
            src = "%s/autoload.js?bokeh-autoload-element=%s&bokeh-app-path=/foo/bar/sliders" % \
                  ("http://localhost:8081/foo/bar/sliders", divid)
            self.assertDictEqual({ 'data-bokeh-doc-id' : '',
                                   'data-bokeh-model-id' : '',
                                   'id' : divid,
                                   'src' : src },
                                 attrs)

    def test_script_attrs_arguments_provided(self):
        r = embed.server_document(arguments=dict(foo=10))
        self.assertTrue('foo=10' in r)
        html = bs4.BeautifulSoup(r, "lxml")
        scripts = html.findAll(name='script')
        self.assertEqual(len(scripts), 1)
        attrs = scripts[0].attrs
        self.assertTrue(set(attrs), set([
            'src',
            'data-bokeh-doc-id',
            'data-bokeh-model-id',
            'id'
        ]))
        divid = attrs['id']
        src = "%s/autoload.js?bokeh-autoload-element=%s&bokeh-absolute-url=%s&foo=10" % \
              ("http://localhost:5006", divid, "http://localhost:5006")
        self.assertDictEqual({ 'data-bokeh-doc-id' : '',
                               'data-bokeh-model-id' : '',
                               'id' : divid,
                               'src' : src },
                             attrs)


@mock.patch('bokeh.document.document.check_integrity')
def test_modelindocument_validates_document_by_default(check_integrity):
    p = figure()
    with embed._ModelInDocument([p]):
        pass
    assert check_integrity.called

@mock.patch('bokeh.document.document.check_integrity')
def test_modelindocument_doesnt_validate_doc_due_to_env_var(check_integrity, monkeypatch):
    monkeypatch.setenv("BOKEH_VALIDATE_DOC", "false")
    p = figure()
    with embed._ModelInDocument([p]):
        pass
    assert not check_integrity.called

@mock.patch('bokeh.document.document.check_integrity')
def test_modelinemptydocument_validates_document_by_default(check_integrity):
    p = figure()
    with embed._ModelInEmptyDocument(p):
        pass
    assert check_integrity.called

@mock.patch('bokeh.document.document.check_integrity')
def test_modelinemptydocument_doesnt_validate_document_due_to_env_var(check_integrity, monkeypatch):
    monkeypatch.setenv("BOKEH_VALIDATE_DOC", "false")
    p = figure()
    with embed._ModelInEmptyDocument(p):
        pass
    assert not check_integrity.called



""" The data in airports.csv and routes.csv is a subset (limited to US airports)
of data available from OpenFlights.org. The complete data was collected on
September 07, 2017 and is available from:

.. code-block:: none

    https://openflights.org/data.html

"""
from __future__ import absolute_import

from bokeh.util.dependencies import import_required
pd = import_required('pandas',
              'airports sample data requires Pandas (http://pandas.pydata.org) to be installed')

import os

from . import _data_dir

airports = pd.read_csv(os.path.join(_data_dir(), 'airports.csv'))
routes = pd.read_csv(os.path.join(_data_dir(), 'routes.csv'))



from mock import patch
import pytest

import bokeh.core.has_props as hp

from bokeh.core.properties import Int, String, NumberSpec, List, Override, Either
from bokeh.core.property.descriptors import BasicPropertyDescriptor, DataSpecPropertyDescriptor

class Parent(hp.HasProps):
    int1 = Int(default=10)
    ds1 = NumberSpec()
    lst1 = List(String)

class Child(Parent):
    int2 = Int()
    str2 = String(default="foo")
    ds2 = NumberSpec()
    lst2 = List(Int, default=[1,2,3])

class OverrideChild(Parent):
    int1 = Override(default=20)

def test_HasProps_default_init():
    p = Parent()
    assert p.int1 == 10
    assert p.ds1 == None
    assert p.lst1 == []

    c = Child()
    assert c.int1 == 10
    assert c.ds1 == None
    assert c.lst1 == []
    assert c.int2 == None
    assert c.str2 == "foo"
    assert c.ds2 == None
    assert c.lst2 == [1,2,3]

def test_HasProps_kw_init():
    p = Parent(int1=30, ds1="foo")
    assert p.int1 == 30
    assert p.ds1 == "foo"
    assert p.lst1 == []

    c = Child(str2="bar", lst2=[2,3,4], ds2=10)
    assert c.int1 == 10
    assert c.ds1 == None
    assert c.lst1 == []
    assert c.int2 == None
    assert c.str2 == "bar"
    assert c.ds2 == 10
    assert c.lst2 == [2,3,4]

def test_HasProps_override():
    ov = OverrideChild()
    assert ov.int1 == 20
    assert ov.ds1 == None
    assert ov.lst1 == []

def test_HasProps_equals():
    p1 = Parent()
    p2 = Parent()
    assert p1.equals(p2)
    p1.int1 = 25
    assert not p1.equals(p2)
    p2.int1 = 25
    assert p1.equals(p2)

def test_HasProps_update():
    c = Child()
    c.update(**dict(lst2=[1,2], str2="baz", int1=25, ds1=dict(field="foo")))
    assert c.int1 == 25
    assert c.ds1 == dict(field="foo")
    assert c.lst1 == []
    assert c.int2 == None
    assert c.str2 == "baz"
    assert c.ds2 ==  None
    assert c.lst2 == [1,2]

def test_HasProps_set_from_json():
    c = Child()
    c.set_from_json('lst2', [1,2])
    assert c.int1 == 10
    assert c.ds1 == None
    assert c.lst1 == []
    assert c.int2 == None
    assert c.str2 == "foo"
    assert c.ds2 ==  None
    assert c.lst2 == [1,2]

    c.set_from_json('ds1', "foo")
    assert c.int1 == 10
    assert c.ds1 == "foo"
    assert c.lst1 == []
    assert c.int2 == None
    assert c.str2 == "foo"
    assert c.ds2 ==  None
    assert c.lst2 == [1,2]

    c.set_from_json('int2', 100)
    assert c.int1 == 10
    assert c.ds1 == "foo"
    assert c.lst1 == []
    assert c.int2 == 100
    assert c.str2 == "foo"
    assert c.ds2 ==  None
    assert c.lst2 == [1,2]


def test_HasProps_update_from_json():
    c = Child()
    c.update_from_json(dict(lst2=[1,2], str2="baz", int1=25, ds1=dict(field="foo")))
    assert c.int1 == 25
    assert c.ds1 == dict(field="foo")
    assert c.lst1 == []
    assert c.int2 == None
    assert c.str2 == "baz"
    assert c.ds2 ==  None
    assert c.lst2 == [1,2]

@patch('bokeh.core.has_props.HasProps.set_from_json')
def test_HasProps_update_from_json_passes_models_and_setter(mock_set):
    c = Child()
    c.update_from_json(dict(lst1=[1,2]), models="foo", setter="bar")
    assert mock_set.called
    assert mock_set.call_args[0] == ('lst1', [1, 2], 'foo', 'bar')
    assert mock_set.call_args[1] == {}

def test_HasProps_set():
    c = Child()
    c.update(**dict(lst2=[1,2], str2="baz", int1=25, ds1=dict(field="foo")))
    assert c.int1 == 25
    assert c.ds1 == dict(field="foo")
    assert c.lst1 == []
    assert c.int2 == None
    assert c.str2 == "baz"
    assert c.ds2 ==  None
    assert c.lst2 == [1,2]

def test_HasProps_set_error():
    c = Child()
    with pytest.raises(AttributeError) as e:
        c.int3 = 10
    assert str(e).endswith("unexpected attribute 'int3' to Child, similar attributes are int2 or int1")
    with pytest.raises(AttributeError) as e:
        c.junkjunk = 10
    assert str(e).endswith("unexpected attribute 'junkjunk' to Child, possible attributes are ds1, ds2, int1, int2, lst1, lst2 or str2")


def test_HasProps_lookup():
    p = Parent()
    d = p.lookup('int1')
    assert isinstance(d, BasicPropertyDescriptor)
    assert d.name == 'int1'
    d = p.lookup('ds1')
    assert isinstance(d, DataSpecPropertyDescriptor)
    assert d.name == 'ds1'
    d = p.lookup('lst1')
    assert isinstance(d, BasicPropertyDescriptor)
    assert d.name == 'lst1'

def test_HasProps_apply_theme():
    c = Child()
    theme = dict(int2=10, lst1=["foo", "bar"])
    c.apply_theme(theme)
    assert c.themed_values() is theme
    c.apply_theme(theme)
    assert c.themed_values() is theme

    assert c.int2 == 10
    assert c.lst1 == ["foo", "bar"]

    assert c.int1 == 10
    assert c.ds1 == None
    assert c.str2 == "foo"
    assert c.ds2 == None
    assert c.lst2 == [1,2,3]

    c.int2 = 25
    assert c.int2 == 25
    assert c.lst1 == ["foo", "bar"]

    assert c.int1 == 10
    assert c.ds1 == None
    assert c.str2 == "foo"
    assert c.ds2 == None
    assert c.lst2 == [1,2,3]

    c.ds2 = "foo"
    assert c.int2 == 25
    assert c.lst1 == ["foo", "bar"]

    assert c.int1 == 10
    assert c.ds1 == None
    assert c.str2 == "foo"
    assert c.ds2 == "foo"
    assert c.lst2 == [1,2,3]

def test_HasProps_unapply_theme():
    c = Child()
    theme = dict(int2=10, lst1=["foo", "bar"])
    c.apply_theme(theme)
    assert c.int2 == 10
    assert c.lst1 == ["foo", "bar"]

    assert c.int1 == 10
    assert c.ds1 == None
    assert c.str2 == "foo"
    assert c.ds2 == None
    assert c.lst2 == [1,2,3]

    c.unapply_theme()
    assert c.int2 == None
    assert c.lst1 == []

    assert c.int1 == 10
    assert c.ds1 == None
    assert c.str2 == "foo"
    assert c.ds2 == None
    assert c.lst2 == [1,2,3]

    assert c.themed_values() == None

class EitherSimpleDefault(hp.HasProps):
    foo = Either(List(Int), Int, default=10)

def test_HasProps_apply_theme_either_simple():

    # check applying multiple themes
    c = EitherSimpleDefault()
    assert c.foo == 10

    theme = dict(foo=20)
    c.apply_theme(theme)
    assert c.foo == 20

    theme = dict(foo=30)
    c.apply_theme(theme)
    assert c.foo == 30

    # check user set before theme
    c = EitherSimpleDefault()
    theme = dict(foo=30)
    c.foo = 50
    c.apply_theme(theme)
    assert c.foo == 50

    # check user set after theme
    c = EitherSimpleDefault()
    theme = dict(foo=30)
    c.apply_theme(theme)
    c.foo = 50
    assert c.foo == 50

    # check user set alt type
    c = EitherSimpleDefault()
    theme = dict(foo=30)
    c.foo = [50]
    c.apply_theme(theme)
    assert c.foo == [50]

    # check themed alt type
    c = EitherSimpleDefault()
    theme = dict(foo=[100])
    c.apply_theme(theme)
    assert c.foo == [100]

class EitherContainerDefault(hp.HasProps):
    foo = Either(List(Int), Int, default=[10])

def test_HasProps_apply_theme_either_container():

    # check applying multiple themes
    c = EitherContainerDefault()
    assert c.foo == [10]

    theme = dict(foo=[20])
    c.apply_theme(theme)
    assert c.foo == [20]

    theme = dict(foo=[30])
    c.apply_theme(theme)
    assert c.foo == [30]

    # check user set before theme
    c = EitherContainerDefault()
    theme = dict(foo=[30])
    c.foo = [50]
    c.apply_theme(theme)
    assert c.foo == [50]

    # check user set after theme
    c = EitherContainerDefault()
    theme = dict(foo=[30])
    c.apply_theme(theme)
    c.foo = [50]
    assert c.foo == [50]

    # check user set alt type
    c = EitherContainerDefault()
    theme = dict(foo=[30])
    c.foo = 50
    c.apply_theme(theme)
    assert c.foo == 50

    # check themed alt type
    c = EitherContainerDefault()
    theme = dict(foo=100)
    c.apply_theme(theme)
    assert c.foo == 100

class IntFuncDefault(hp.HasProps):
    foo = Int(default=lambda: 10)

def test_HasProps_apply_theme_func_default():

    # check applying multiple themes
    c = IntFuncDefault()
    assert c.foo == 10

    theme = dict(foo=20)
    c.apply_theme(theme)
    assert c.foo == 20

    theme = dict(foo=30)
    c.apply_theme(theme)
    assert c.foo == 30

    # check user set before theme
    c = IntFuncDefault()
    theme = dict(foo=30)
    c.foo = 50
    c.apply_theme(theme)
    assert c.foo == 50

    # check user set after theme
    c = IntFuncDefault()
    theme = dict(foo=30)
    c.apply_theme(theme)
    c.foo = 50
    assert c.foo == 50

# TODO (bev) need to skip for now
# def test_HasProps_pretty():
#     p = Parent()
#     assert p.pretty() == "bokeh.core.tests.test_has_props.Parent(ds1=None, int1=10, lst1=[])"
#     old = hp.IPython
#     hp.IPython = False
#     with pytest.raises(RuntimeError):
#         p.pretty()
#     hp.IPython = old

def test_HasProps_pprint(capsys):
    p = Parent()
    p.pprint()
    out, err = capsys.readouterr()
    assert out == "bokeh.core.tests.test_has_props.Parent(ds1=None, int1=10, lst1=[])\n"
    assert err == ""



from __future__ import absolute_import, print_function

def test_dir():
    import bokeh
    names = dir(bokeh)
    assert "__version__" in names
    assert "__base_version__" in names
    assert "license" in names
    assert "sampledata" in names

def test_version_defined():
    import bokeh
    assert bokeh.__version__ != 'unknown'

# This is failing only on Travis for some reason
# def test_license():
#     import sys
#     import bokeh
#     try:
#         bokeh.license()
#     except Exception as e:
#         print("If LICENSE.txt does not exist in bokeh/ subdir, one way to fix this may be to run 'python setup.py develop'", file=sys.stderr)
#         raise e



""" The data in airports.json is a subset of US airports with field
elevations > 1500 meters. The query result was taken from

.. code-block:: none

    http://services.nationalmap.gov/arcgis/rest/services/GlobalMap/GlobalMapWFS/MapServer/10/query

on October 15, 2015.
"""
from __future__ import absolute_import

from bokeh.util.dependencies import import_required
pd = import_required('pandas',
              'airports sample data requires Pandas (http://pandas.pydata.org) to be installed')

import json
import os

from . import _data_dir

with open(os.path.join(_data_dir(), 'airports.json'), 'r') as data_file:
    content = data_file.read()
    airports = json.loads(content)
    schema = [['attributes', 'nam'], ['attributes', 'zv3'], ['geometry', 'x'], ['geometry', 'y']]
    data = pd.io.json.json_normalize(airports['features'], meta=schema)
    data.rename(columns={'attributes.nam': 'name', 'attributes.zv3': 'elevation'}, inplace=True)
    data.rename(columns={'geometry.x': 'x', 'geometry.y': 'y'}, inplace=True)



import bokeh.document.util as bdu

from bokeh.document import Document

from .setup import AnotherModelInTestDocument, SomeModelInTestDocument

def test_compute_one_attribute_patch():
    d = Document()
    root1 = SomeModelInTestDocument(foo=42)
    child1 = SomeModelInTestDocument(foo=43)
    root1.child = child1
    d.add_root(root1)

    before = d.to_json()

    root1.foo = 47

    after = d.to_json()

    patch = bdu.compute_patch_between_json(before, after)

    expected = dict(references=[],
                    events=[
                        {'attr': u'foo',
                         'kind': 'ModelChanged',
                         'model': {'id': None,
                                   'type': 'SomeModelInTestDocument'},
                         'new': 47}
                    ])
    expected['events'][0]['model']['id'] = root1._id
    assert expected == patch

    d2 = Document.from_json(before)
    d2.apply_json_patch(patch)
    assert root1.foo == d2.roots[0].foo

def test_compute_two_attribute_patch():
    d = Document()
    root1 = SomeModelInTestDocument(foo=42)
    child1 = AnotherModelInTestDocument(bar=43)
    root1.child = child1
    d.add_root(root1)

    before = d.to_json()

    root1.foo=47
    child1.bar=57

    after = d.to_json()

    patch = bdu.compute_patch_between_json(before, after)

    expected = dict(references=[],
                    events=[
                        {'attr': u'bar',
                         'kind': 'ModelChanged',
                         'model': {'id': None,
                                   'type': 'AnotherModelInTestDocument'},
                         'new': 57},
                        {'attr': u'foo',
                         'kind': 'ModelChanged',
                         'model': {'id': None,
                                   'type': 'SomeModelInTestDocument'},
                         'new': 47}
                        ])
    expected['events'][0]['model']['id'] = child1._id
    expected['events'][1]['model']['id'] = root1._id

    # order is undefined, so fix our expectation if needed
    assert len(patch['events']) == 2
    if patch['events'][0]['model']['type'] == 'AnotherModelInTestDocument':
        pass
    else:
        tmp = expected['events'][0]
        expected['events'][0] = expected['events'][1]
        expected['events'][1] = tmp

    assert expected == patch

    d2 = Document.from_json(before)
    d2.apply_json_patch(patch)
    assert root1.foo == d2.roots[0].foo
    assert root1.child.bar == d2.roots[0].child.bar

def test_compute_remove_root_patch():
    d = Document()
    root1 = SomeModelInTestDocument(foo=42)
    child1 = AnotherModelInTestDocument(bar=43)
    root1.child = child1
    d.add_root(root1)

    before = d.to_json()

    d.remove_root(root1)

    after = d.to_json()

    patch = bdu.compute_patch_between_json(before, after)

    expected = dict(references=[],
                    events= [
                        {'kind': 'RootRemoved',
                         'model': {'id': None,
                                   'type': 'SomeModelInTestDocument'}}
                    ])
    expected['events'][0]['model']['id'] = root1._id

    assert expected == patch

    d2 = Document.from_json(before)
    d2.apply_json_patch(patch)
    assert d2.roots == []

def test_compute_add_root_patch():
    d = Document()
    root1 = SomeModelInTestDocument(foo=42)
    child1 = AnotherModelInTestDocument(bar=43)
    root1.child = child1
    d.add_root(root1)

    before = d.to_json()

    root2 = SomeModelInTestDocument(foo=57)
    d.add_root(root2)

    after = d.to_json()

    patch = bdu.compute_patch_between_json(before, after)

    expected = {
        'references' : [
            { 'attributes': {'child': None, 'foo': 57},
              'id': None,
              'type': 'SomeModelInTestDocument'}
        ],
        'events' : [
            { 'kind': 'RootAdded',
              'model': {'id': None,
                        'type': 'SomeModelInTestDocument'}
            }
        ]
    }

    expected['references'][0]['id'] = root2._id
    expected['events'][0]['model']['id'] = root2._id

    assert expected == patch

    d2 = Document.from_json(before)
    d2.apply_json_patch(patch)
    assert len(d2.roots) == 2
    assert d2.roots[0].foo == 42
    assert d2.roots[1].foo == 57



''' Provides the ``ApplicationContext`` class.

'''
from __future__ import absolute_import

import logging
log = logging.getLogger(__name__)

from tornado import gen

from .session import ServerSession

from ..application.application import ServerContext, SessionContext
from ..document import Document
from ..protocol.exceptions import ProtocolError
from ..util.tornado import _CallbackGroup, yield_for_all_futures

class _RequestProxy(object):
    def __init__(self, request):
        args_copy = dict(request.arguments)
        if 'bokeh-protocol-version' in args_copy: del args_copy['bokeh-protocol-version']
        if 'bokeh-session-id' in args_copy: del args_copy['bokeh-session-id']
        self._args = args_copy
    @property
    def arguments(self):
        return self._args

class BokehServerContext(ServerContext):
    def __init__(self, application_context):
        self.application_context = application_context
        self._callbacks = _CallbackGroup(self.application_context.io_loop)

    def _remove_all_callbacks(self):
        self._callbacks.remove_all_callbacks()

    @property
    def sessions(self):
        result = []
        for session in self.application_context.sessions:
            result.append(session.session_context)
        return result

    def add_next_tick_callback(self, callback):
        self._callbacks.add_next_tick_callback(callback)

    def remove_next_tick_callback(self, callback):
        self._callbacks.remove_next_tick_callback(callback)

    def add_timeout_callback(self, callback, timeout_milliseconds):
        self._callbacks.add_timeout_callback(callback, timeout_milliseconds)

    def remove_timeout_callback(self, callback):
        self._callbacks.remove_timeout_callback(callback)

    def add_periodic_callback(self, callback, period_milliseconds):
        self._callbacks.add_periodic_callback(callback, period_milliseconds)

    def remove_periodic_callback(self, callback):
        self._callbacks.remove_periodic_callback(callback)

class BokehSessionContext(SessionContext):
    def __init__(self, session_id, server_context, document):
        self._document = document
        self._session = None
        super(BokehSessionContext, self).__init__(server_context,
                                                  session_id)
        # request arguments used to instantiate this session
        self._request = None

    def _set_session(self, session):
        self._session = session

    @gen.coroutine
    def with_locked_document(self, func):
        if self._session is None:
            # this means we are in on_session_created, so no locking yet,
            # we have exclusive access
            yield yield_for_all_futures(func(self._document))
        else:
            self._session.with_document_locked(func, self._document)

    @property
    def destroyed(self):
        if self._session is None:
            # this means we are in on_session_created
            return False
        else:
            return self._session.destroyed

    @property
    def request(self):
        return self._request


class ApplicationContext(object):
    ''' Server-side holder for bokeh.application.Application plus any associated data.
        This holds data that's global to all sessions, while ServerSession holds
        data specific to an "instance" of the application.
    '''

    def __init__(self, application, io_loop=None, url=None):
        self._application = application
        self._loop = io_loop
        self._sessions = dict()
        self._pending_sessions = dict()
        self._session_contexts = dict()
        self._server_context = None
        self._url = url

    @property
    def io_loop(self):
        return self._loop

    @property
    def application(self):
        return self._application

    @property
    def url(self):
        return self._url

    @property
    def server_context(self):
        if self._server_context is None:
            self._server_context = BokehServerContext(self)
        return self._server_context

    @property
    def sessions(self):
        return self._sessions.values()

    def run_load_hook(self):
        try:
            result = self._application.on_server_loaded(self.server_context)
            if isinstance(result, gen.Future):
                log.error("on_server_loaded returned a Future; this doesn't make sense "
                          "because we run this hook before starting the IO loop.")
        except Exception as e:
            log.error("Error in server loaded hook %r", e, exc_info=True)

    def run_unload_hook(self):
        try:
            result = self._application.on_server_unloaded(self.server_context)
            if isinstance(result, gen.Future):
                log.error("on_server_unloaded returned a Future; this doesn't make sense "
                          "because we stop the IO loop right away after calling on_server_unloaded.")
        except Exception as e:
            log.error("Error in server unloaded hook %r", e, exc_info=True)

        self.server_context._remove_all_callbacks()

    @gen.coroutine
    def create_session_if_needed(self, session_id, request=None):
        # this is because empty session_ids would be "falsey" and
        # potentially open up a way for clients to confuse us
        if len(session_id) == 0:
            raise ProtocolError("Session ID must not be empty")

        if session_id not in self._sessions and \
           session_id not in self._pending_sessions:
            future = self._pending_sessions[session_id] = gen.Future()

            doc = Document()


            session_context = BokehSessionContext(session_id,
                                                  self.server_context,
                                                  doc)
            # using private attr so users only have access to a read-only property
            session_context._request = _RequestProxy(request)

            # expose the session context to the document
            # use the _attribute to set the public property .session_context
            doc._session_context = session_context


            try:
                yield yield_for_all_futures(self._application.on_session_created(session_context))
            except Exception as e:
                log.error("Failed to run session creation hooks %r", e, exc_info=True)

            self._application.initialize_document(doc)

            session = ServerSession(session_id, doc, io_loop=self._loop)
            del self._pending_sessions[session_id]
            self._sessions[session_id] = session
            session_context._set_session(session)
            self._session_contexts[session_id] = session_context

            # notify anyone waiting on the pending session
            future.set_result(session)

        if session_id in self._pending_sessions:
            # another create_session_if_needed is working on
            # creating this session
            session = yield self._pending_sessions[session_id]
        else:
            session = self._sessions[session_id]

        raise gen.Return(session)

    def get_session(self, session_id):
        if session_id in self._sessions:
            session = self._sessions[session_id]
            return session
        else:
            raise ProtocolError("No such session " + session_id)

    @gen.coroutine
    def _discard_session(self, session, should_discard):
        if session.connection_count > 0:
            raise RuntimeError("Should not be discarding a session with open connections")
        log.debug("Discarding session %r last in use %r milliseconds ago", session.id, session.milliseconds_since_last_unsubscribe)

        session_context = self._session_contexts[session.id]

        # session.destroy() wants the document lock so it can shut down the document
        # callbacks.
        def do_discard():
            # while we yielded for the document lock, the discard-worthiness of the
            # session may have changed.
            # However, since we have the document lock, our own lock will cause the
            # block count to be 1. If there's any other block count besides our own,
            # we want to skip session destruction though.
            if should_discard(session) and session.expiration_blocked_count == 1:
                session.destroy()
                del self._sessions[session.id]
                del self._session_contexts[session.id]
            else:
                log.debug("Session %r was scheduled to discard but came back to life", session.id)
        yield session.with_document_locked(do_discard)

        # session lifecycle hooks are supposed to be called outside the document lock,
        # we only run these if we actually ended up destroying the session.
        if session_context.destroyed:
            try:
                result = self._application.on_session_destroyed(session_context)
                yield yield_for_all_futures(result)
            except Exception as e:
                log.error("Failed to run session destroy hooks %r", e, exc_info=True)

        raise gen.Return(None)

    @gen.coroutine
    def _cleanup_sessions(self, unused_session_linger_milliseconds):
        def should_discard_ignoring_block(session):
            return session.connection_count == 0 and \
                (session.milliseconds_since_last_unsubscribe > unused_session_linger_milliseconds or \
                 session.expiration_requested)
        # build a temp list to avoid trouble from self._sessions changes
        to_discard = []
        for session in self._sessions.values():
            if should_discard_ignoring_block(session) and not session.expiration_blocked:
                to_discard.append(session)
        # asynchronously reconsider each session
        for session in to_discard:
            if should_discard_ignoring_block(session) and not session.expiration_blocked:
                yield self._discard_session(session, should_discard_ignoring_block)

        raise gen.Return(None)



''' Internal utility functions used by ``bokeh.client``

'''
from __future__ import absolute_import, print_function

def server_url_for_websocket_url(url):
    ''' Convert an ``ws(s)`` URL for a Bokeh server into the appropriate
    ``http(s)`` URL for the websocket endpoint.

    Args:
        url (str):
            An ``ws(s)`` URL ending in ``/ws``

    Returns:
        str:
            The corresponding ``http(s)`` URL.

    Raises:
        ValueError:
            If the input URL is not of the proper form.

    '''
    if url.startswith("ws:"):
        reprotocoled = "http" + url[2:]
    elif url.startswith("wss:"):
        reprotocoled = "https" + url[3:]
    else:
        raise ValueError("URL has non-websocket protocol " + url)
    if not reprotocoled.endswith("/ws"):
        raise ValueError("websocket URL does not end in /ws")
    return reprotocoled[:-2]

def websocket_url_for_server_url(url):
    ''' Convert an ``http(s)`` URL for a Bokeh server websocket endpoint into
    the appropriate ``ws(s)`` URL

    Args:
        url (str):
            An ``http(s)`` URL

    Returns:
        str:
            The corresponding ``ws(s)`` URL ending in ``/ws``

    Raises:
        ValueError:
            If the input URL is not of the proper form.

    '''
    if url.startswith("http:"):
        reprotocoled = "ws" + url[4:]
    elif url.startswith("https:"):
        reprotocoled = "wss" + url[5:]
    else:
        raise ValueError("URL has unknown protocol " + url)
    if reprotocoled.endswith("/"):
        return reprotocoled + "ws"
    else:
        return reprotocoled + "/ws"



from __future__ import absolute_import

from bokeh.util.dependencies import import_required
pd = import_required('pandas',
              'glucose sample data requires Pandas (http://pandas.pydata.org) to be installed')

from os.path import join

from . import _data_dir

data_dir = _data_dir()

data = pd.read_csv(
    join(data_dir, 'CGM.csv'),
    sep=',',
    parse_dates=[1],
    index_col=1
)



"""

"""
from __future__ import absolute_import

from bokeh.util.dependencies import import_required

pd = import_required('pandas', 'commits sample data requires Pandas (http://pandas.pydata.org) to be installed')

from os.path import dirname, join

data = pd.read_csv(join(dirname(__file__), 'commits.txt.gz'),
                   parse_dates=True, header=None, names=['day', 'datetime'], index_col='datetime')

data = data.tz_localize('GMT').tz_convert('US/Central')
data['time'] = data.index.time



import pytest

import bokeh.client.util as bcu

def test_server_url_for_websocket_url_with_ws():
    assert bcu.server_url_for_websocket_url("ws://foo.com/ws") == "http://foo.com/"

def test_server_url_for_websocket_url_with_wss():
    assert bcu.server_url_for_websocket_url("wss://foo.com/ws") == "https://foo.com/"

def test_server_url_for_websocket_url_bad_proto():
    with pytest.raises(ValueError):
        bcu.server_url_for_websocket_url("junk://foo.com/ws")

def test_server_url_for_websocket_url_bad_ending():
    with pytest.raises(ValueError):
        bcu.server_url_for_websocket_url("ws://foo.com/junk")
    with pytest.raises(ValueError):
        bcu.server_url_for_websocket_url("wss://foo.com/junk")

def test_websocket_url_for_server_url_with_http():
    assert bcu.websocket_url_for_server_url("http://foo.com") == "ws://foo.com/ws"
    assert bcu.websocket_url_for_server_url("http://foo.com/") == "ws://foo.com/ws"

def test_websocket_url_for_server_url_with_https():
    assert bcu.websocket_url_for_server_url("https://foo.com") == "wss://foo.com/ws"
    assert bcu.websocket_url_for_server_url("https://foo.com/") == "wss://foo.com/ws"

def test_websocket_url_for_server_url_bad_proto():
    with pytest.raises(ValueError):
        bcu.websocket_url_for_server_url("junk://foo.com")
